/*--------------------------------------------------------------*/
/* gvwrap.c:							*/
/*	Tcl module generator for gvtk				*/
/*--------------------------------------------------------------*/

#ifdef TCL_WRAPPER

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <tk.h>

#include "gvtk.h"

Tcl_Interp *gvinterp;

/*------------------------------------------------------*/
/* Create command callback functions in Tcl/Tk    	*/    
/*------------------------------------------------------*/    

typedef struct {
   const char	*cmdstr;
   void		(*func)();
} cmdstruct;

static cmdstruct gv_commands[] =
{
   {"config", (void *)gvtk_config},
   {"start", (void *)gvtk_start},
   {"render", (void *)gvtk_render},
   {"page", (void *)gvtk_page},
   {"document", (void *)gvtk_document},
   {"zoom", (void *)gvtk_zoom},
   {"rotate", (void *)gvtk_rotate},
   {"offset", (void *)gvtk_offset},
   {"tag", (void *)gvtk_tag},
   {"quit", (void *)gvtk_quit},
   {"", NULL}  /* sentinel */
};

/*--------------------------------------------------------------*/
/* gvtk commands						*/
/*--------------------------------------------------------------*/

/*--------------------------------------------------------------*/
/* function: gvtk_config					*/
/* command:  config 						*/
/*--------------------------------------------------------------*/

int gvtk_config(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   static char *subCmds[] = {"window", "id", "pixmap", "windownames", "remove", NULL};
   enum SubIdx {WindowIdx, IDIdx, PixmapIdx, WindowNamesIdx, RemoveIdx};
   int idx, result;
   Tk_Window tkwind;
   GVWindowData *newwin, *winptr, *tmpwin;

   if (objc == 1) {
      Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
      return TCL_ERROR;
   }
   else if ((result = Tcl_GetIndexFromObj(interp, objv[1],
	(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK) {
      return result;
   }
   switch(idx) {
       case WindowIdx:
	  if (objc == 2) {
             if (areawin == NULL)
		Tcl_SetResult(interp, "There are no target windows.", NULL);
	     else if (areawin->area == (Tk_Window)NULL)
		Tcl_SetResult(interp, "No current window set.", NULL);
	     else
                Tcl_AppendElement(interp, Tk_PathName(areawin->area));
	  }
	  else if (objc == 3) {
	     tkwind = Tk_NameToWindow(interp, Tcl_GetString(objv[2]), gvobjs.topwin);
	     if (tkwind == (Tk_Window)NULL) {
		Tcl_SetResult(interp, "No such window!", NULL);
		return TCL_ERROR;
	     }
	     /* This either creates or sets focus to a window */
	     newwin = new_ghostscript_target(interp, tkwind);
	     if (newwin == NULL)
		return TCL_ERROR;
	     else
		areawin = newwin;

	     if (areawin->gs_channel == (Tcl_Channel)NULL)
		ghostinit(tkwind);
	  }
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;

       case RemoveIdx:
	  /* Kill the interpreter and remove the window from the list */

	  if (objc == 2) {
             exit_gs(areawin, False);
	     newwin = NULL;
	     for (winptr = gvobjs.windowlist; winptr != NULL; ) {
		if (winptr == areawin) {
		   if (newwin == NULL)
		      gvobjs.windowlist = winptr->next;
		   else
		      newwin->next = winptr->next;
	 	   Tk_FreeGC(dpy, winptr->gc);
		   Tk_FreePixmap(dpy, winptr->dbuf);
		   tmpwin = winptr->next;
		   free(winptr);
		   winptr = tmpwin;
		}
		else {
		   newwin = winptr;
		   winptr = winptr->next;
		}
	     }
	  }
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;

       case IDIdx:
	  if (objc == 2)
             Tcl_SetObjResult(interp, Tcl_NewIntObj(areawin->window));
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;

       case PixmapIdx:
	  if (objc == 2)
             Tcl_SetObjResult(interp, Tcl_NewIntObj(areawin->dbuf));
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;

       case WindowNamesIdx:
	  if (objc == 2) {
	     for (winptr = gvobjs.windowlist; winptr != NULL; winptr = winptr->next)
                Tcl_AppendElement(interp, Tk_PathName(winptr->area));
	  }
	  else {
	     Tcl_WrongNumArgs(interp, 2, objv, "(no arguments)");
	     return TCL_ERROR;
	  }
	  break;
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_render					*/
/* command:  render <filename>					*/
/*								*/
/* Loads a file, parses its DSC contents, and attaches it to	*/
/* the current focus window.					*/
/*--------------------------------------------------------------*/

int gvtk_render(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   FileList *newfptr;

   if (objc == 1) {
      if (areawin->fileptr != NULL)
	 Tcl_SetResult(interp, areawin->fileptr->filename, NULL);
   }
   else if (objc != 2) {
      Tcl_WrongNumArgs(interp, 1, objv, "?filename?");
      return TCL_ERROR;
   }
   else {
      newfptr = new_file(Tcl_GetString(objv[1]));
      if (newfptr != NULL) {
         if (areawin->fileptr != NULL) {
	    if (areawin->fileptr != newfptr) {
	       areawin->fileptr = NULL;
	       exit_gs(areawin, True);	// Restart ghostscript
	    }
         }
         areawin->fileptr = newfptr;
         areawin->page = 0;
         renderpage(areawin);
      }
      else {
	 Tcl_AppendResult(interp, "Could not open file ", Tcl_GetString(objv[1]), NULL);
         return TCL_ERROR;
      }
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_page						*/
/* command:  page [<number>] [<option>]				*/
/*								*/
/*	Where options are:  goto, mark, label, orient, etc.	*/
/*	Only "mark" takes an additional optional argument, the	*/
/* 	boolean value "true" or "false".			*/
/*	"page" with no options returns the current page.	*/
/*	"page <number>" is equivalent to "page <number> goto".	*/
/*--------------------------------------------------------------*/

int gvtk_page(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   static char *subCmds[] = {"goto", "label", "orient", "media", "size", "bbox",
		"mark", NULL};
   enum SubIdx {GoToIdx, LabelIdx, OrientIdx, MediaIdx, SizeIdx, BBoxIdx, MarkIdx};
   int idx, result, pageno, cmdidx;
   struct page *curpage;
   Tcl_Obj *lobj;

   if (objc == 1) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(1 + areawin->page));
      return TCL_OK;
   }
   else {
      result = Tcl_GetIntFromObj(interp, objv[1], &pageno);
      if (result != TCL_OK) {
	 Tcl_ResetResult(interp);
	 pageno = areawin->page;
         cmdidx = 1;
      }
      else {
	 objc--;	// objc will count the # arguments but not the page #.
	 pageno--;	// Internally, page count starts at zero.
	 cmdidx = 2;
      }

      if (objc < cmdidx)
	 idx = GoToIdx;
      else {
         result = Tcl_GetIndexFromObj(interp, objv[cmdidx], (CONST84 char **)subCmds,
			"option", 0, &idx);
         if (result != TCL_OK) return result;
      }
   }

   /* Check for valid page */

   if (!areawin->fileptr || !areawin->fileptr->doc) {
      Tcl_SetResult(interp, "Document contains no page information", NULL);
      return TCL_ERROR;
   }
   else if (pageno >= areawin->fileptr->doc->numpages) {
      Tcl_SetResult(interp, "Not that many pages in document", NULL);
      return TCL_ERROR;
   }
   else if (pageno < 0) {
      Tcl_SetResult(interp, "Page out of range", NULL);
      return TCL_ERROR;
   }

   curpage = &areawin->fileptr->doc->pages[pageno];

   /* Check for valid number of arguments for each option */

   switch(idx) {
      case MarkIdx:
	 if ((objc != 2) && (objc != 3)) {
	    Tcl_WrongNumArgs(interp, cmdidx, objv, "?true|false?");
	    return TCL_ERROR;
	 }
	 break;
      case GoToIdx:
	 if ((objc != 1) && (objc != 2)) {
	    Tcl_WrongNumArgs(interp, cmdidx, objv, "?page?");
	    return TCL_ERROR;
	 }
	 break;
      default:
	 if (objc != 2) {
	    Tcl_WrongNumArgs(interp, cmdidx, objv, "(no arguments)");
	    return TCL_ERROR;
	 }
	 break;
   }

   /* Process the option */

   switch(idx) {
       case GoToIdx:
	  areawin->page = pageno;
	  renderpage(areawin);
	  break;

       case LabelIdx:
	  Tcl_SetObjResult(interp, Tcl_NewStringObj(curpage->label, -1));
	  break;

       case OrientIdx:
	  if (curpage->orientation == 1) 
	     Tcl_SetResult(interp, "portrait", NULL);
	  else if (curpage->orientation == 2) 
	     Tcl_SetResult(interp, "landscape", NULL);
	  else
	     Tcl_SetResult(interp, "unknown", NULL);
	  break;

       case MediaIdx:
	  if (curpage->media)
	     Tcl_SetObjResult(interp, Tcl_NewStringObj(curpage->media->name, -1));
	  else
	     Tcl_SetResult(interp, "none", NULL);
	  break;

       case SizeIdx:
          lobj = Tcl_NewListObj(0, NULL);
	  if (curpage->media) {
             Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj(curpage->media->width));
             Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj(curpage->media->height));
	  }
	  else {
             Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj(curpage->boundingbox[2] -
			curpage->boundingbox[0]));
             Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj(curpage->boundingbox[3] -
			curpage->boundingbox[1]));
 	  }
          Tcl_SetObjResult(interp, lobj);
	  break;

       case BBoxIdx:
          lobj = Tcl_NewListObj(0, NULL);
          Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj((double)curpage->boundingbox[0]));
          Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj((double)curpage->boundingbox[1]));
          Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj((double)curpage->boundingbox[2]));
          Tcl_ListObjAppendElement(interp, lobj,
			Tcl_NewIntObj((double)curpage->boundingbox[3]));
          Tcl_SetObjResult(interp, lobj);
	  break;

      case MarkIdx:
	 if (objc == 3) {
	    int pmark;
	    result = Tcl_GetBooleanFromObj(interp, objv[cmdidx + 1], &pmark);
	    if (result != TCL_OK) return result;
	    curpage->marked = (Boolean)pmark;
	 }
	 else
	    Tcl_SetObjResult(interp, Tcl_NewBooleanObj(curpage->marked));
	 break;
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_document					*/
/* command:  document <option>					*/
/*--------------------------------------------------------------*/

int gvtk_document(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   static char *subCmds[] = {"title", "date", "pages", "encapsulated", "order", NULL};
   enum SubIdx {TitleIdx, DateIdx, PagesIdx, EPSFIdx, OrderIdx};
   int idx, result;
   struct document *curdoc;

   if (!areawin->fileptr) {
      Tcl_SetResult(interp, "No file", NULL);
      return TCL_ERROR;
   }
   else
      curdoc = areawin->fileptr->doc;
 
   if (!curdoc) {
      Tcl_SetResult(interp, "Document has no structure", NULL);
      return TCL_ERROR;
   }
   
   if (objc == 1) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(1 + areawin->page));
      return TCL_OK;
   }
   else if ((result = Tcl_GetIndexFromObj(interp, objv[1],
	(CONST84 char **)subCmds, "option", 0, &idx)) != TCL_OK) {
      return result;
   }

   switch(idx) {
      case TitleIdx:
	 Tcl_SetObjResult(interp, Tcl_NewStringObj(curdoc->title, -1));
	 break;

      case DateIdx:
	 Tcl_SetObjResult(interp, Tcl_NewStringObj(curdoc->date, -1));
	 break;

      case EPSFIdx:
	 Tcl_SetObjResult(interp, Tcl_NewBooleanObj(curdoc->epsf));
	 break;

      case PagesIdx:
	 Tcl_SetObjResult(interp, Tcl_NewIntObj(curdoc->numpages));
	 break;

      case OrderIdx:
	 switch (curdoc->pageorder) {
	    case 3:
	       Tcl_SetResult(interp, "ascend", NULL);
	       break;
	    case 4:
	       Tcl_SetResult(interp, "descend", NULL);
	       break;
	    case 5:
	       Tcl_SetResult(interp, "special", NULL);
	       break;
	    default:
	       Tcl_SetResult(interp, "none", NULL);
	       break;
	 }
	 break;
   }

   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_quit						*/
/* command:  quit						*/
/*--------------------------------------------------------------*/

int gvtk_quit(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   GVWindowData *checkwin;
   FileListPtr frecord;

   /* Close all open files */
   for (frecord = gvobjs.filelist; frecord != NULL; frecord = frecord->next) {
      if (frecord->fps != (FILE *)NULL) {
	 fclose(frecord->fps);
	 frecord->fps = (FILE *)NULL;
      }
   }

   if (objc != 1) {
      Tcl_WrongNumArgs(interp, 1, objv, "option ?arg ...?");
      return TCL_ERROR;
   }

   /* Exit all interpreters */
   for (checkwin = gvobjs.windowlist; checkwin != NULL; checkwin = checkwin->next)
      exit_gs(checkwin, False);

   Tcl_Exit(0);
   return TCL_OK;	/* Not reached */
}

/*--------------------------------------------------------------*/
/* function: gvtk_rotate					*/
/* command:  rotate <amount>					*/
/*--------------------------------------------------------------*/

int gvtk_rotate(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   int rotval, result;

   if (objc == 1) {
      Tcl_SetObjResult(interp, Tcl_NewIntObj(areawin->rotate));
      result = TCL_OK;
   }
   else {
      result = Tcl_GetIntFromObj(interp, objv[1], &rotval);   
      if (result == TCL_OK) {
         areawin->rotate = rotval;
         renderpage(areawin);
      }
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_zoom						*/
/* command:  zoom <amount>					*/
/*--------------------------------------------------------------*/

int gvtk_zoom(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   double zoomval;
   int result;

   if (objc == 1) {
      Tcl_SetObjResult(interp, Tcl_NewDoubleObj((double)areawin->zoom));
      result = TCL_OK;
   }
   else {
      result = Tcl_GetDoubleFromObj(interp, objv[1], &zoomval);   
      if (result == TCL_OK) {
         areawin->zoom = (float)zoomval;
         renderpage(areawin);
      }
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* function: gvtk_offset					*/
/* command:  offset <xamount yamount> [copy]			*/
/*								*/
/* Without argument "copy", sets the page offset from the	*/
/* bottom left corner (in pixels).  With argument "copy", 	*/
/* makes a fast copy of the page at the indicated offset.	*/
/*--------------------------------------------------------------*/

int gvtk_offset(ClientData clientData, Tcl_Interp *interp,
        int objc, Tcl_Obj *CONST objv[])
{
   int llen, result;
   int offsetx, offsety;
   Tcl_Obj *xylist, *lobj;

   if (objc == 1) {
      xylist = Tcl_NewListObj(0, NULL);
      Tcl_ListObjAppendElement(interp, xylist, Tcl_NewIntObj(areawin->offsetx));
      Tcl_ListObjAppendElement(interp, xylist, Tcl_NewIntObj(areawin->offsety));
      Tcl_SetObjResult(interp, xylist);
      result = TCL_OK;
   }
   else {
      if (objc == 2) {
         result = Tcl_ListObjLength(interp, objv[1], &llen);   
	 if (result != TCL_OK) return result;

	 if (llen != 2) {
	    Tcl_SetResult(interp, "list must contain x y positions", NULL);
	    return TCL_ERROR;
	 }
	 else {
	    result = Tcl_ListObjIndex(interp, objv[1], 0, &lobj);
	    if (result != TCL_OK) return result;
	    result = Tcl_GetIntFromObj(interp, lobj, &offsetx);
	    if (result != TCL_OK) return result;
	    result = Tcl_ListObjIndex(interp, objv[1], 1, &lobj);
	    if (result != TCL_OK) return result;
	    result = Tcl_GetIntFromObj(interp, lobj, &offsety);
	    if (result != TCL_OK) return result;
	 }
      }
      else if (objc >= 3) {
	 result = Tcl_GetIntFromObj(interp, objv[1], &offsetx);
	 if (result != TCL_OK) return result;
	 result = Tcl_GetIntFromObj(interp, objv[2], &offsety);
	 if (result != TCL_OK) return result;
      }

      if (objc > 3) {
	 if (!strncasecmp(Tcl_GetString(objv[3]), "copy", 4)) {
	    copyoffset(areawin, offsetx, offsety);
	 }
	 else {
	    Tcl_SetResult(interp, "unknown option", NULL);
	    return TCL_ERROR;
	 }
      }
      else {
         areawin->offsetx = offsetx;
         areawin->offsety = offsety;
         renderpage(areawin);
      }
   }
   return GvTagCallback(interp, objc, objv);
}

/*--------------------------------------------------------------*/
/* Initialization procedure for Tcl/Tk				*/
/*--------------------------------------------------------------*/

int
Gvtk_Init(Tcl_Interp *interp)
{
   char command[256];
   int cmdidx;
   Tk_Window tktop;
   char version_string[20];

   /* Interpreter sanity checks */
   if (interp == NULL) return TCL_ERROR;

   /* Remember the interpreter */
   gvinterp = interp;

   if (Tcl_InitStubs(interp, "8.1", 0) == NULL) return TCL_ERROR;
   
   strcpy(command, "gvtk::");

   /* Create the start command */

   tktop = Tk_MainWindow(interp);

   /* Create all of the commands (except "simple") */

   for (cmdidx = 0; gv_commands[cmdidx].func != NULL; cmdidx++) {
      sprintf(command + 6, "%s", gv_commands[cmdidx].cmdstr);
      Tcl_CreateObjCommand(interp, command, 
		(Tcl_ObjCmdProc *)gv_commands[cmdidx].func,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);
   }

   /* Command which creates a "simple" window (this is a top-	*/
   /* level command, not one in the gvtk namespace which	*/
   /* is why I treat it separately from the other commands).	*/

   Tcl_CreateObjCommand(interp, "simple",
		(Tcl_ObjCmdProc *)Tk_SimpleObjCmd,
		(ClientData)tktop, (Tcl_CmdDeleteProc *) NULL);

   /* Set $GVTK_VERSION and $GVTK_REVISION as Tcl variables */

   sprintf(version_string, "%d", PROG_REVISION);
   Tcl_SetVar(interp, "GVTK_REVISION", version_string, TCL_GLOBAL_ONLY);

   sprintf(version_string, "%g", PROG_VERSION);
   Tcl_SetVar(interp, "GVTK_VERSION", version_string, TCL_GLOBAL_ONLY);

   /* Export the namespace commands */

   Tcl_Eval(interp, "namespace eval gvtk namespace export *");
   Tcl_PkgProvide(interp, "Gvtk", version_string);

   Tcl_InitHashTable(&GvTagTable, TCL_STRING_KEYS);
   return TCL_OK;
}

#endif /* TCL_WRAPPER */
